import numpy as np
import pandas as pd
import altair as alt

def detect_outliers(s, width=250, height=250):
    """
    Detects outliers in a pandas series
    Returns a threshold value for the lower bound and upper bound of the outliers and Plot a violin plot of the observations
    Parameters
    ----------
    s : pandas.core.series.Series
        Pandas Series for which the outliers need to be found
    width: int, default = 250
        The width of the plot. Default set to 250
    height: int, default = 250
        The height of the plot. Default set to 250

    Returns
    -------
    List of integers
        Boolean array with same length as the input,
        indices of outlier marked.
    `altair plot`
        An interactive altair correlation plot
    Examples
    --------
    >>> from snapedautility.detect_outliers import detect_outliers 
    >>> s = pd.Series([1,1,2,3,4,5,6,9,10,13,40])
    >>> detect_outliers(s)
    [-8.0 , 20.0],  alt.Chart(...) 
    """
    if not isinstance(s, pd.Series):
        raise TypeError("s should be a pandas series")
    if s.shape[0] < 1:
        raise ValueError("s should have at least one element")
    if (not isinstance(height, int)) or (not isinstance(width, int)):
        raise TypeError("height and width should be integers")
        
    q1 = s.quantile(0.25)
    q3 = s.quantile(0.75)
    factor = 1.5
    inter_quantile_range = q3 - q1

    df = s.rename("values").to_frame()

    # Box Plot
    plt = alt.Chart(df, title="Box Plot with Outliers").mark_boxplot().encode(
        x='values'
    ).properties(width=width, height=height)

    return [q1 - factor * inter_quantile_range, q3 + factor * inter_quantile_range], plt
    


import pandas as pd
import altair as alt
import numpy as np


def plot_corr(df, features=None, width=250, height=250):
    """
    Generates the pearson correlation plot for a list of numeric features in a given dataframe.

    Parameters
    ----------
    data : pandas.core.frame.DataFrame
        The input dataframe
    features : list
        List of feature names as string
        len(features) >=2
        None returns plot of all numeric features
    width: int, default = 250
        The width of the plot. Default set to 250
    height: int, default = 250
        The height of the plot. Default set to 250

    Returns
    -------
    `altair plot`
        An interactive altair correlation plot
    Examples
    --------
    >>> from snapedautility.plot_corr import plot_corr
    >>> df = penguins_data
    >>> plot_corr(df, ["Culmen Length (mm)", "Culmen Depth (mm)", 'Species'])
    """
    # Testing whether the input data frame is pd.DataFrame type
    if isinstance(df, pd.DataFrame) == False:
        raise TypeError("The thing you passed in `df` is not a valid Pandas DataFrame")

    # Tests whether input features is of the type list
    if features is not None:
        if not isinstance(features, list):
            raise TypeError("Please pass in a list for `features`")

    # Testing whether input features has at least two features
    if features is not None:
        if len(features) < 2:
            raise ValueError("At least two features should be selected")

    # Subsetting numerical features from the input dataframe
    if features is None:
        if df.select_dtypes(include="number").shape[1] < 2:
            raise ValueError("Dataframe should have at least two numerical features")
        else:
            df = df.select_dtypes(include="number")
    else:
        if df[features].select_dtypes(include="number").shape[1] < 2:
            raise ValueError(
                "The feature list you entered should have at least two numerical features"
            )
        else:
            df = df[features].select_dtypes(include="number")

    # Creating corr_df dataframe
    corr_df = df.corr().stack().reset_index(name="corr")
    corr_df["abs"] = corr_df["corr"].abs()

    # Correlation plot
    corr_plot = (
        alt.Chart(corr_df, title="Correlation Plot for Numerical Features")
        .mark_rect()
        .encode(
            x=alt.X("level_0", title="Numerical Features"),
            y=alt.Y("level_1", title="Numerical Features"),
            color=alt.Color(
                "corr",
                title="Correlation",
                scale=alt.Scale(domain=(-1, 1),scheme="purpleorange"),
            ),
            tooltip=alt.Tooltip("corr"),
        )
        .properties(width=width, height=height)
    )

    return corr_plot



import pandas as pd
import numpy as np
import altair as alt


def plot_histograms(df, features, facet_columns=3, width=125, height=125):
    """
    Plots histogram given numeric features of the input dataframe, and
    plots bar charts for categorical features of the input dataframe

    Parameters
    ----------
    df : pandas.core.frame.DataFrame
        Input dataframe
    features : list
        List of feature names as string
    facet_columns : int
        Number of columns in Facet options.
    width: int
        The width of sub-plot for each feature. Default set to 125
    height: int
        The height of sub-plot for each feature Default set to 125

    Returns
    -------
    `altair plot`
        Returns altair plot

    Examples
    --------
    >>> from snapedautility.plot_histograms import plot_histograms
    >>> df = penguins_data
    >>> plot_histograms(df, ["species", "bill_length_mm", "island"], 100, 100)
    """
    features_set = set(features)

    # Some basic validation on the function's input parameters.
    if not isinstance(df, pd.DataFrame):
        raise ValueError("The data must be in type of Pandas DataFrame.")
    elif ((len(features) == 0) or (features_set.issubset(set(df.columns)) == False)):
        raise ValueError("All features must exist in the columns of the input Pandas DataFrame.")
    elif (not isinstance(width, int)) or (not isinstance(height, int)) or (height <= 0) or (width <= 0):
        raise ValueError("Width and height of the plot must be a positive integer")

    # Select categorical columns and numeric columns
    cat_cols = list(set(df.select_dtypes(include=["object"]).columns).intersection(features_set))
    numeric_cols = list(set(df.select_dtypes(include=np.number).columns).intersection(features_set))

    # Create alt.Chart for categorical features
    categorical_barplot = (
        alt.Chart(df)
        .mark_bar()
        .encode(
            x=alt.X(alt.repeat(), type='nominal'),
            y=alt.Y('count()', title="Count of Records"))
        .properties(width=width, height=height)
        .repeat(cat_cols, columns=facet_columns)
    )

    # Create alt.Chart for numeric features.
    numeric_barplot = (
        alt.Chart(df)
        .mark_bar()
        .encode(
            x=alt.X(alt.repeat(), type='quantitative', bin=alt.Bin(maxbins=30)),
            y=alt.Y('count()', title="Numberic Features"))
        .properties(width=width, height=height)
        .repeat(numeric_cols, columns=facet_columns)
    )

    #categorical_barplot
    histograms_plot = categorical_barplot & numeric_barplot
    histograms_plot.title = "Histograms for Specified Features"
    return histograms_plot